
.wy-nav-content {
  background: white;
}

.wy-nav-content h1,
.wy-nav-content h2,
.wy-nav-content h3,
.wy-nav-content h4,
.wy-nav-content h5,
.wy-nav-content h6 {
  font-family: Arial, Helvetica, sans-serif;
  color: rgb(153,51,51); /* mit red */
}

.wy-nav-content a:not(.image) {
  color: rgb(153,51,51); /* mit red */
}

.wy-nav-content b { color: rgb(153,51,51); /* mit red */ }

.wy-nav-content i { color: rgb(153,51,51); /* mit red */ }

.wy-nav-content em { color: rgb(153,51,51); /* mit red */ }

.wy-nav-content code { color: rgb(153,51,51); /* mit red */ }

.wy-side-nav-search { background-color: rgb(153,51,51); /* mit red */ }

.sig.sig-object.py {
  background-color: rgb(242,242,242) !important; /* light gray */
  border-top-color: rgb(153,51,51) !important; /* mit red */
}

.sig-paren {
  color: rgb(153,51,51); /* mit red */
}